import * as readline from "n-readlines";
import * as nativePathFromPath from "path";
import * as fs from "fs";
import {DebuggerVariable, Attribute, VariableType} from "./debugger";

const nativePath = {
    resolve: function (...args: string[]): string {
        const nat = nativePathFromPath.resolve(...args);
        if (process.platform === "win32" && this.cobcpath === "docker" && this.gdbpath === "docker") {
            return nat.replace(/.*:/, s => "/" + s.toLowerCase().replace(":", "")).replace(/\\/g, "/");
        }
        return nat;
    },
    basename: function (path: string): string {
        return nativePathFromPath.basename(path);
    },
    isAbsolute: function (path: string): boolean {
        return nativePathFromPath.isAbsolute(path);
    },
    join: function (...args: string[]) {
        return nativePathFromPath.join(...args);
    }
};

const procedureRegex = /\/\*\sLine:\s([0-9]+)(\s+:\sEntry\s)?/i;
const procedureFixRegex = /#line\s([0-9]+)\s".*\.c"/i;
const attributeRegex = /static\sconst\scob_field_attr\s(a_[0-9]+).*\{(0x\d+),\s*([0-9-]*),\s*([0-9-]*),\s*(0x\d{4}),.*/i;
const dataStorageRegex = /static\s+(.*)\s+(b_[0-9]+)(\;|\[\d+\]).*\/\*\s+([0-9a-z_\- ]+)\s+\*\//i;
const fieldRegex = /static\s+cob_field\s+([0-9a-z_]+)\s+\=\s+\{(\d+)\,\s+([0-9a-z_]+).+\&(a_\d+).*\/\*\s+([0-9a-z_\-]+)\s+\*\//i;
const fileIncludeRegex = /#include\s+\"([0-9a-z_\-\.\s]+)\"/i;
const fileCobolRegex = /\/\*\sGenerated from\s+([0-9a-z_\-\/\.\s\\:]+)\s+\*\//i;
const functionRegex = /\/\*\sProgram\slocal\svariables\sfor\s'(.*)'\s\*\//i;
const versionRegex = /\/\*\sGenerated by\s+cobc\s([0-9a-z\-\.]+)\s+\*\//i;
const subroutineRegex = /\sPerform\s/i;
const frame_ptrFindRegex = /frame\_ptr\-\-\;/;
const fixOlderFormat = /cob\_trace\_stmt/;

globalThis.varOccurs = [];

export class Line {
    fileCobol: string;
    fileC: string;
    lineCobol: number;
    lineC: number;
    endPerformLine: number;   // 002 - stepOver in routines with "perform"

    public constructor(filePathCobol: string, lineCobol: number, filePathC: string, lineC: number) {
        this.fileCobol = filePathCobol;
        this.lineCobol = lineCobol;
        this.fileC = filePathC;
        this.lineC = lineC;
        this.endPerformLine = -1;          // 002 - stepOver in routines with "perform"
    }

    public toString(): string {
        return `${this.fileCobol}:${this.lineCobol} > ${this.fileC}:${this.lineC}`;
    }
}

export class SourceMap {
    private cwd: string;
    private lines: Line[] = new Array<Line>();
    private variablesByCobol = new Map<string, DebuggerVariable>();
    private variablesByC = new Map<string, DebuggerVariable>();
    private attributes = new Map<string, Attribute>();
    private dataStorages = new Map<string, DebuggerVariable>();
    private version: string;
    private lineBefore: string = "";
    private performLine: number = -1; // 002 - stepOver in routines with "perform" 
    private isVersion2_2_or_3_1_1: boolean = false;

    constructor(cwd: string, filesCobol: string[]) {
        this.cwd = fs.realpathSync(nativePathFromPath.resolve(cwd));
        filesCobol.forEach(e => {
            this.parse(nativePath.basename(e.split('.').slice(0, -1).join('.') + '.c'));
        });
    }

    private parse(fileC: string): void {
        let nat = fileC;
        if (!nativePath.isAbsolute(fileC)) {
            nat = nativePathFromPath.resolve(this.cwd, fileC);
            fileC = nativePath.resolve(this.cwd, fileC);
        }

        const basename = nativePath.basename(fileC);
        const cleanedFile = basename.substring(0, basename.lastIndexOf(".c"));

        let lineNumber = 0;

        const reader = new readline(process.platform === "win32" ? nat : fileC);
        let row: false | Buffer;
        let fileCobol: string;
        let functionName: string;
        while (row = reader.next()) {
            const line = row.toString();
            let match = fileCobolRegex.exec(line);
            if (match) {
                if (!nativePath.isAbsolute(match[1])) {
                    fileCobol = nativePath.resolve(this.cwd, match[1]);
                } else {
                    fileCobol = match[1];
                }
            }
            match = functionRegex.exec(line);
            if (match) {
                functionName = match[1].toLowerCase() + "_";
            }
            match = procedureRegex.exec(line);
            if (match && !match[2]) {
                if (this.lines.length > 0 && fileNameCompare(this.lines[this.lines.length - 1].fileCobol, fileCobol) && this.lines[this.lines.length - 1].lineCobol === parseInt(match[1])) {
                    this.lines.pop();
                }
                if(subroutineRegex.exec(line)) 
                    this.performLine=-2; // must find line of frame_ptr
                else 
                    this.performLine=-1; 
                this.lines.push(new Line(fileCobol, parseInt(match[1]), fileC, lineNumber + 2));
            }
            // fix new codegen
            match = procedureFixRegex.exec(line);
            if (match && this.lines.length > 0) {
                let isOldFormat = fixOlderFormat.exec(this.lineBefore);
                if(this.isVersion2_2_or_3_1_1 || !isOldFormat){ // Is it in the old format?
                    let line = this.lines.pop();
                    line.lineC = parseInt(match[1]);
                    this.lines.push(line);
                }
            }
            match = attributeRegex.exec(line);
            if (match) {
                const attribute = new Attribute(match[1], VariableType[match[2]], parseInt(match[3]), parseInt(match[4]), match[5]);
                this.attributes.set(`${cleanedFile}.${match[1]}`, attribute);
            }
            match = dataStorageRegex.exec(line);
            if (match) {
                let size: number = null;
                if (match[3].startsWith("[")) {
                    size = parseInt(match[3].substring(1, match[3].length - 1));
                }
                const dataStorage = new DebuggerVariable(match[4], match[2], functionName, new Attribute(null, VariableType[match[1]], 0, 0), size);
                this.dataStorages.set(`${functionName}.${dataStorage.cName}`, dataStorage);
                this.variablesByC.set(`${functionName}.${dataStorage.cName}`, dataStorage);
                this.variablesByCobol.set(`${functionName}.${dataStorage.cobolName.toUpperCase()}`, dataStorage);
            }
            match = fieldRegex.exec(line);
            if (match) {
                const attribute = this.attributes.get(`${cleanedFile}.${match[4]}`);
                const dataStorage = this.dataStorages.get(`${functionName}.${match[3]}`);
                const field = new DebuggerVariable(match[5], match[1], functionName, attribute, parseInt(match[2]));

                this.variablesByC.set(`${functionName}.${field.cName}`, field);

                if (dataStorage) {
                    dataStorage.addChild(field);
                    this.variablesByCobol.set(`${functionName}.${dataStorage.cobolName.toUpperCase()}.${field.cobolName.toUpperCase()}`, field);
                } else {
                    this.variablesByCobol.set(`${functionName}.${field.cobolName.toUpperCase()}`, field);
                }
            }
            match = fileIncludeRegex.exec(line);
            if (match) {
                this.parse(match[1]);
            }
            match = versionRegex.exec(line);
            if (match) {
                this.version = match[1];
                if(this.version.startsWith("2.2") || this.version.startsWith("3.1.1")) this.isVersion2_2_or_3_1_1 = true;
            }
            if(this.performLine == -2){
                match = frame_ptrFindRegex.exec(line);
                if(match){
                    let line = this.lines.pop();
                    line.endPerformLine = lineNumber+1;
                    this.lines.push(line);
                }
            }            
            this.lineBefore = line;
            lineNumber++;
        }
    }

    public getVariablesByC(): IterableIterator<DebuggerVariable> {
        return this.variablesByC.values();
    }

    public getVariablesByCobol(): IterableIterator<DebuggerVariable> {
        return this.variablesByCobol.values();
    }

    public getLinesCount(): number {
        return this.lines.length;
    }

    public getVariablesCount(): number {
        return this.variablesByC.size;
    }

    public getVariableByC(varC: string): DebuggerVariable {
        if (this.variablesByC.has(varC)) {
            return this.variablesByC.get(varC);
        }
        return null;
    }

    public findVariableByCobol(functionName: string, name: string): DebuggerVariable {
        for (const key of this.variablesByCobol.keys()) {
            if (key.startsWith(functionName) && key.endsWith(`.${name.toUpperCase()}`)) {
                return this.variablesByCobol.get(key);
            }
        }
        return null;
    }

    public getVariableByCobol(path: string): DebuggerVariable {
        return this.variablesByCobol.get(path);
    }

    public hasLineCobol(fileC: string, lineC: number): boolean {
        if(!fileC || !lineC) return false;
        if (!nativePath.isAbsolute(fileC)) {
            fileC = nativePath.join(this.cwd, fileC);
        }
        return this.lines.some(e => e.fileC === fileC && e.lineC === lineC);
    }

    // 002 - stepOver in routines with "perform"
    public hasLineSubroutine(fileC: string, lineC: number): number {
        if(!fileC || !lineC) return -1;
        if (!nativePath.isAbsolute(fileC)) {
            fileC = nativePath.join(this.cwd, fileC);
        }
        let line = this.lines.find(e => e.fileC === fileC && e.lineC === lineC) ?? new Line('', 0, '', 0);
        return line.endPerformLine;
    }
    // 002

    public hasLineC(fileCobol: string, lineCobol: number): boolean {
        if (!nativePath.isAbsolute(fileCobol)) {
            fileCobol = nativePath.join(this.cwd, fileCobol);
        }
        return this.lines.some(e => fileNameCompare(e.fileCobol, fileCobol) && e.lineCobol === lineCobol);
    }

    public getLineC(fileCobol: string, lineCobol: number): Line {
        if (!nativePath.isAbsolute(fileCobol)) {
            fileCobol = nativePath.join(this.cwd, fileCobol);
        }
        return this.lines.find(e => fileNameCompare(e.fileCobol, fileCobol) && e.lineCobol === lineCobol) ?? new Line('', 0, '', 0);
    }

    public getLineCobol(fileC: string, lineC: number): Line {
        if (!nativePath.isAbsolute(fileC)) {
            fileC = nativePath.join(this.cwd, fileC);
        }
        return this.lines.find(e => e.fileC === fileC && e.lineC === lineC) ?? new Line('', 0, '', 0);
    }

    public getVersion(): string {
        return this.version;
    }

    public toString(): string {
        let out = `SourceMap created: lines ${this.lines.length}, vars ${this.variablesByC.size}\n`;

        this.lines.forEach(e => {
            out += e.toString() + "\n";
        });

        this.variablesByC.forEach((value, key) => {
            out += `${key}[${value.attribute?.cName}] > ${value.cobolName}\n`;
        });

        this.variablesByCobol.forEach((value, key) => {
            out += `${key}[${value.attribute?.cName}] > ${value.cName}\n`;
        });

        return out;
    }

}

function fileNameCompare(fileNameOne: string, fileNameTwo: string): boolean {
    if(process.platform === "win32")
        return fileNameOne.toUpperCase() === fileNameTwo.toUpperCase();
    else
        return fileNameOne === fileNameTwo;
}
